#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <signal.h>
#include <assert.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <linux/if.h>
#include <linux/if_tun.h>
#include <linux/if_bridge.h>

#include <uuid/uuid.h>

#include "xenner.h"

/* ------------------------------------------------------------------ */

static struct xenvm xen = {
    .pg_total = MB_TO_PG(128),
    .debug    = 1,
    .corefd   = -1,
    .name     = "no-name",
    .vcpus    = 1,
};

/* ------------------------------------------------------------------ */

#define MONITOR_GREETING  "xenner " VERSION " monitor\r\n"
#define MONITOR_ERROR     "unknown command\r\n"
#define MONITOR_PROMPT    "(qemu) "
#define MONITOR_HELP      \
    "help             - this text\r\n" \
    "system_shutdown  - gracefull guest shutdown\r\n" \
    "quit             - destoy guest and quit\r\n"

static int xwrite(int fd, const void *buf, size_t count)
{
    int rc, try;

    for (try = 0; try < 5; try++) {
	rc = write(fd, buf, count);
	if (rc == count)
	    break;
	if (rc == -1 && EINTR == errno)
	    continue;
	fprintf(stderr, "%s: write(mon, buf, %zd) = %d, errno %s\n",
		__FUNCTION__, count, rc, strerror(errno));
    }
    return rc;
}

static int qemu_monitor_line(struct xenvm *xen, int fd, char *line)
{
    static const char nl[]     = "\r\n";
    static const char prompt[] = MONITOR_PROMPT;
    static const char error[]  = MONITOR_ERROR;
    static const char help[]   = MONITOR_HELP;

    d0printf("%s: \"%s\"\n", __FUNCTION__, line);
    xwrite(fd, nl, strlen(nl));

    if (0 == strcmp(line, "help")) {
	xwrite(fd, help, strlen(help));

    } else if (0 == strcmp(line, "system_shutdown")) {
	kill(getpid(), SIGINT);

    } else if (0 == strcmp(line, "quit")) {
	kill(getpid(), SIGTERM);

    } else if (strlen(line)) {
	xwrite(fd, error, strlen(error));
 
    }
    xwrite(fd, prompt, strlen(prompt));
    return 0;
}

static int qemu_monitor_data(struct xenvm *xen, int fd, void *data)
{
    static const char msg[] = MONITOR_GREETING MONITOR_PROMPT;
    char buf[256];
    int rc, pos, i;

    rc = read(fd, buf, sizeof(buf));
    switch (rc) {
    case -1:
	if (EIO == errno) {
	    if (!xen->mon_connected) {
		d1printf("%s: disconnected\n", __FUNCTION__);
		xen->mon_connected = 0;
	    }
	    return 0;
	} else if (EAGAIN == errno) {
	    if (!xen->mon_connected) {
		d1printf("%s: connected\n", __FUNCTION__);
		xwrite(fd, msg, strlen(msg));
	    }
	    xen->mon_connected = 1;
	    return 0;
	} else {
	    d1printf("%s: read: %s\n", __FUNCTION__, strerror(errno));
	}
	break;
    case 0:
	d1printf("%s: read: EOF\n", __FUNCTION__);
	break;
    default:
	buf[rc] = 0;
	xwrite(fd, buf, rc); /* echo */
	pos  = strlen(xen->mon_buffer);
	pos += snprintf(xen->mon_buffer+pos,
			sizeof(xen->mon_buffer)-pos, "%s", buf);

	/* try split into lines, failing that just return and wait for
	   more data  (unless someone overflows our buffer) */
	i = 0;
	while (i < pos) {
	    buf[i] = xen->mon_buffer[i];
	    if (buf[i] == '\r' || buf[i] == '\n') {
		buf[i] = '\0';
		qemu_monitor_line(xen, fd, buf);
		memmove(xen->mon_buffer, xen->mon_buffer+i+1, pos-i);
		pos -= i+1;
		i = 0;
		continue;
	    }
	    i++;
	}
	if (pos == sizeof(xen->mon_buffer)) {
	    d0printf("%s: zap flooded buffer\n", __FUNCTION__);
	    xen->mon_buffer[0] = '\0';
	}
	break;
    }
    return 0;
}

static int qemu_monitor_config(struct xenvm *xen, char *arg)
{
    char *dev = NULL;
    int tty = -1;

    d1printf("%s: %s\n", __FUNCTION__, arg);
    if (0 == strcmp(arg, "pty")) {
	tty = getpt();
	if (-1 == tty)
	    d1printf("%s: getpt: %s\n", __FUNCTION__, strerror(errno));
	if (-1 == grantpt(tty))
	    d1printf("%s: grantpt: %s\n", __FUNCTION__, strerror(errno));
	if (-1 == unlockpt(tty))
	    d1printf("%s: unlockpt: %s\n", __FUNCTION__, strerror(errno));
	dev = ptsname(tty);
	/* make read() return -EIO when unconnected, so we can
	 * figure when someone (libvirt ...) connects */
	close(open(dev, O_RDWR));
    }
    if (dev) {
	struct termios tio;
	
	d1printf("%s: %s\n", __FUNCTION__, dev);
	tcgetattr (tty, &tio);
	tio.c_lflag &= ~(ECHO|ICANON|ISIG);
	tio.c_cc[VMIN] = 1;
	tio.c_cc[VTIME] = 0;
	tcsetattr (tty, TCSAFLUSH, &tio);
	/* libvirt expects this on stderr */
	fprintf(stderr, "char device redirected to %s\n", dev);
    }
    return tty;
}

static int qemu_monitor_setup(struct xenvm *xen, int tty)
{
    xenner_register_cb(xen, XENNER_CB_IOFD, "monitor",
		       tty, NULL, qemu_monitor_data);
    qemu_monitor_data(xen, tty, NULL);
    return 0;
}

/* ------------------------------------------------------------------ */

static int tap_fd_to_name(int fd, char *buf, int len)
{
    /* try to figure interface name from file handle */
    struct ifreq tun, sys;
    int sock = -1, ifn, match = 0;

    /* get tun dev mac address */
    memset(&tun, 0, sizeof(tun));
    if (-1 == ioctl(fd, SIOCGIFHWADDR, &tun))
	goto out;

    /*
     * Loop over interfaces to find the interface name with the
     * given mac address.  Uses last match, to skip bridge interfaces
     * which might have the same mac address.
     */
    sock = socket(PF_INET, SOCK_STREAM, 0);
    if (sock < 0)
	goto out;
    for (ifn = 0; ifn < 65536; ifn++) {
	memset(&sys, 0, sizeof(sys));
	sys.ifr_ifindex = ifn;
	if (ioctl(sock, SIOCGIFNAME, &sys) < 0) {
	    if (ENODEV == errno)
		continue;
	    goto out;
	}
	if (ioctl(sock, SIOCGIFHWADDR, &sys) < 0)
	    goto out;
	if (0 != memcmp(tun.ifr_hwaddr.sa_data, sys.ifr_hwaddr.sa_data, 6))
	    continue;
	/* matched */
	strncpy(buf, sys.ifr_name, len);
	match++;
    }

    if (match) {
	/* make persist and close, so netbackd can open it by name */
	ioctl(fd, TUNSETPERSIST, 1);
	close(fd);
	goto out;
    }

out:
    if (-1 != sock)
	close(sock);
    return match;
}

static int qemu_parse_string(char *arg, char *tag, char *buf, int len)
{
    char *pos;
    int i = 0;

    pos = strstr(arg, tag);
    if (!pos)
	return -1;
    pos += strlen(tag);
    while (*pos != '\0' && *pos != ',' &&
	   *pos != '\r' && *pos != '\n' &&
	   i < len-1) {
	buf[i] = *pos;
	i++; pos++;
    }
    buf[i] = 0;
    return i;
}

static int qemu_parse_int(char *arg, char *tag, int def)
{
    char buf[32];

    if (qemu_parse_string(arg, tag, buf, sizeof(buf)) <= 0)
	return def;
    return atoi(buf);
}

static int qemu_net_config(struct xenvm *xen, char *arg)
{
    char dummy[32];
    unsigned int vlan;
    int fd;
    
    d1printf("%s: %s\n", __FUNCTION__, arg);
    if (0 == strncmp(arg, "none", 4)) {
	return 0;
    }
    if (0 == strncmp(arg, "user", 4)) {
	d1printf("%s: user network not implemented\n", __FUNCTION__);
	return 0;
    }
    
    vlan = qemu_parse_int(arg, "vlan=", 0);
    if (vlan >= sizeof(xen->nics)/sizeof(xen->nics[0])) {
	d1printf("%s: vlan number too big (%u, %zd max)\n", __FUNCTION__,
		 vlan, sizeof(xen->nics)/sizeof(xen->nics[0])-1);
	return 0;
    }
    xen->nics[vlan].configured++;
    
    if (0 == strncmp(arg, "tap", 3)) {
	fd = qemu_parse_int(arg, "fd=", -1);
	qemu_parse_string(arg, "ifname=", xen->nics[vlan].ifname,
			  sizeof(xen->nics[vlan].ifname));
	qemu_parse_string(arg, "script=", dummy, sizeof(dummy));
	if (-1 != fd)
	    tap_fd_to_name(fd, xen->nics[vlan].ifname,
			   sizeof(xen->nics[vlan].ifname));
	d1printf("%s: %d: if=\"%s\"\n", __FUNCTION__, vlan,
		 xen->nics[vlan].ifname);
    }

    if (0 == strncmp(arg, "nic", 3)) {
	qemu_parse_string(arg, "macaddr=", xen->nics[vlan].mac,
			  sizeof(xen->nics[vlan].mac));
	qemu_parse_string(arg, "model=", dummy, sizeof(dummy));
	d1printf("%s: %d: mac=\"%s\"\n", __FUNCTION__, vlan,
		 xen->nics[vlan].mac);
    }
    
    return 0;
}

/* ------------------------------------------------------------------ */

static int qemu_vnc_config(struct xenvm *xen, char *arg)
{
    d1printf("%s: %s\n", __FUNCTION__, arg);
    xen->qemu_vnc = arg;
    return 0;
}

static void prepare_environment(struct xenvm *xen, char *argv0)
{
    char *path[] = {
	".",
	".",
	"/usr/" LIB "/xenner",
	"/usr/local/" LIB "/xenner",
    };
    char cwd[256];
    char exe[256];
    char buf[256];
    char *h;
    int i;

    getcwd(cwd, sizeof(cwd));
    path[0] = cwd;

    strncpy(exe, argv0, sizeof(exe));
    h = strrchr(exe,'/');
    if (h)
	*h = 0;
    path[1] = exe;

    for (i = 0; i < sizeof(path)/sizeof(path[0]); i++) {
	snprintf(buf, sizeof(buf), "%s/%s", path[i], "libxenctrl.so.3.0");
	if (0 != access(buf, R_OK))
	    continue;
	setenv("LD_LIBRARY_PATH", path[i], 1);
	d1printf("%s: LD_LIBRARY_PATH=%s\n", __FUNCTION__, path[i]);
	break;
    }
}

static int qemu_dm_setup(struct xenvm *xen, char *argv0)
{
    char domid[16];
    char *argv[] = {
	"/usr/" LIB "/xen/bin/qemu-dm",
	"-M", "xenpv",
	"-vnc", xen->qemu_vnc,
	"-d", domid,
	"-domain-name", xen->name,
	NULL
    };

    if (!xen->qemu_vnc)
	return 0;

    snprintf(domid, sizeof(domid), "%d", xen->domid);
    prepare_environment(xen, argv0);
    xen->qemu_dm = fork();
    switch (xen->qemu_dm) {
    case -1: /* Oops */
	d0printf("%s: fork: %s\n", __FUNCTION__, strerror(errno));
	return -1;
    case 0: /* child */
	execv(argv[0], argv);
	d1printf("execv(%s): %s\n", argv[0], strerror(errno));
	exit(1);
    default: /* parent */
	return 0;
    }
}

int qemu_cleanup(struct xenvm *xen)
{
    if (xen->qemu_dm > 0)
	kill(xen->qemu_dm, SIGTERM);
    return 0;
}

/* ------------------------------------------------------------------ */

static void openlog(struct xenvm *xen, char *filename)
{
    xen->logfile = fopen(filename, "w");
    if (!xen->logfile) {
	logprintf(xen, "open %s: %s\n", filename, strerror(errno));
	return;
    }
    setvbuf(xen->logfile, NULL, _IONBF, 0);
    logprintf(xen, "logging to %s\n", filename);
}

static void usage(FILE *fp)
{
    fprintf(fp,
	    "QEMU PC emulator version 0.9.0\n"
	    "/*\n"
	    " * no, it isn't, this just makes libvirt qemu version detection happy,\n"
	    " * so you can manage xenner guests using libvirts qemu driver\n"
	    " */\n"
	    "\n"
	    "xenner -- run xen kernels in kvm\n"
	    "\n"
	    "usage: xenner [options]\n"
	    "options (qemu compatible):\n"
	    "   -help               print this text\n"
	    "   -kernel <kernel>    xenified kernel image\n"
	    "   -initrd <ramdisk>   ramdisk\n"
	    "   -append <append>    kernel command line\n"
	    "   -m      <memory>    vm memory in megabytes\n"
	    "   -name   <string>    domain name\n"
	    "   -hda    <file>      disk image\n"
	    "   -cdrom  <file>      cdrom image\n"
	    "   -net    <config>    qemu style network setup\n"
	    "   -vnc    <display>   vnc display\n"
	    "\n"
	    "options (xenner only):\n"
	    "   -network <mac>,<br> xenner style network setup\n"
	    "   -debug  <nr>        set debug level\n"
	    "   -logfile <file>     set logfile\n"
	    "   -nostderr           send messages to logfile only\n"
	    "\n"
	    "some more qemu-ish options are accepted but ignored.\n"
	    "\n"
	    "-- \n"
	    "(c) 2007 Gerd Hoffmann <kraxel@redhat.com>\n"
	    "\n");
}

int main(int argc, char *argv[])
{
    char *colors = NULL;
    char *uuid = NULL;
    int monitor_fd = -1;
    int opt;

    xen.domid = getpid();
    INIT_LIST_HEAD(&xen.callbacks[XENNER_CB_IOFD]);
    INIT_LIST_HEAD(&xen.callbacks[XENNER_CB_EVTCHN]);

    for (opt = 1; opt < argc; opt++) {
	if (opt < argc-1) {
	    /* options with one argument */
	    if (0 == strcmp(argv[opt], "-boot") ||
		0 == strcmp(argv[opt], "-usbdevice")) {
		/* ignore */
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-M")) {
		/*
		 * libvirt detection: calls us with -M pc
		 *  -> we go logging to a file then as stderr
		 *     isn't visible to the user
		 *  -> ignore otherwise
		 */
		openlog(&xen, "/var/log/xenner.debug");
		xen.nostderr = 1;
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-hda")) {
		xen.disk = argv[opt+1];
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-cdrom")) {
		xen.cdrom = argv[opt+1];
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-m")) {
		xen.pg_total = MB_TO_PG(atoi(argv[opt+1]));
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-smp")) {
#if 0 /* no SMP yet */
		xen.vcpus = atoi(argv[opt+1]);
#endif
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-name")) {
		snprintf(xen.name, sizeof(xen.name), "%s", argv[opt+1]);
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-monitor")) {
		monitor_fd = qemu_monitor_config(&xen, argv[opt+1]);
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-net")) {
		qemu_net_config(&xen, argv[opt+1]);
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-vnc")) {
		qemu_vnc_config(&xen, argv[opt+1]);
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-kernel")) {
		xen.kernel = argv[opt+1];
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-append")) {
		xen.cmdline = argv[opt+1];
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-initrd")) {
		xen.ramdisk = argv[opt+1];
		opt++; continue;

	    /* the ones below are xenner only */
	    } else if (0 == strcmp(argv[opt], "-debug")) {
		xen.debug = atoi(argv[opt+1]);
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-network")) {
		if (2 == sscanf(argv[opt+1], "%31[^,],%31s",
				xen.nics[0].mac,
				xen.nics[0].bridge))
		    xen.nics[0].configured = 1;
		opt++; continue;

	    } else if (0 == strcmp(argv[opt], "-logfile")) {
		openlog(&xen, argv[opt+1]);
		opt++; continue;
	    }
	}
	/* options without arguments */
	if (0 == strcmp(argv[opt], "-usb") ||
	    0 == strcmp(argv[opt], "-no-acpi")) {
	    /* ignore */
	    continue;

	} else if (0 == strcmp(argv[opt], "-nostderr")) {
	    xen.nostderr = 1;
	    continue;

	} else if (0 == strcmp(argv[opt], "-h") ||
		   0 == strcmp(argv[opt], "-help")) {
	    usage(stdout);
	    exit(0);
	}

	logprintf(&xen, "unknown option (or missing arg): %s\n", argv[opt]);
	if (xen.logfile)
	    usage(xen.logfile);
	if (!xen.nostderr)
	    usage(stderr);
	exit(1);
    }

    if (1 == argc) {
	/* no arguments, libvirt expects help text on stdout */
	usage(stdout);
	exit(0);
    }
    if (NULL == xen.kernel) {
	logprintf(&xen, "no kernel specified\n"); 
	exit(1);
    }

    loginit(colors);
    logprintf(&xen, "started as: ");
    for (opt = 0; opt < argc; opt++)
	logprintf(&xen, "\"%s\"%s", argv[opt], (opt == argc-1) ? "\n" : " "); 

    if (NULL == uuid) {
	uuid_t u;
	strcpy(xen.vm, "/vm/");
	uuid_generate(u);
	uuid_unparse(u, xen.vm+4);
    } else {
	sprintf(xen.vm, "/vm/%.36s", uuid);
    }

    /* init kvm */
    if (0 != kvmbits_init(&xen))
	exit(1);

    if (-1 != monitor_fd)
	qemu_monitor_setup(&xen, monitor_fd);
    qemu_dm_setup(&xen, argv[0]);
    
    banner_print(&xen, "boot");
    kvm_run(xen.kvm, 0);
    exit(0);
}
